#!/usr/bin/env python3

#
# NOSA HEADER START
#
# The contents of this file are subject to the terms of the NASA Open
# Source Agreement (NOSA), Version 1.3 only (the "Agreement").  You may
# not use this file except in compliance with the Agreement.
#
# You can obtain a copy of the agreement at
#   docs/NASA_Open_Source_Agreement_1.3.txt
# or
#   https://sscweb.gsfc.nasa.gov/WebServices/NASA_Open_Source_Agreement_1.3.txt.
#
# See the Agreement for the specific language governing permissions
# and limitations under the Agreement.
#
# When distributing Covered Code, include this NOSA HEADER in each
# file and include the Agreement file at
# docs/NASA_Open_Source_Agreement_1.3.txt.  If applicable, add the
# following below this NOSA HEADER, with the fields enclosed by
# brackets "[]" replaced with your own identifying information:
# Portions Copyright [yyyy] [name of copyright owner]
#
# NOSA HEADER END
#
# Copyright (c) 2013-2020 United States Government as represented by
# the National Aeronautics and Space Administration. No copyright is
# claimed in the United States under Title 17, U.S.Code. All Other
# Rights Reserved.
#

"""
Module defining classes to represent the output format option classes from
<https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.<br>

Copyright &copy; 2013-2020 United States Government as represented by the
National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S.Code. All Other Rights Reserved.
"""

import xml.etree.ElementTree as ET
from enum import Enum


class DateFormat(Enum):
    """
    Enumerations representing the DateFormat defined
    in <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
    """
    YYYY_DDD = 'yyyy_ddd'
    YY_MM_DD = 'yy_mm_dd'
    YY_MMM_DD = 'yy_Mmm_dd'
    YY_CMMM_DD = 'yy_CMMM_dd'


class DegreeFormat(Enum):
    """
    Enumerations representing the DegreeFormat defined
    in <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
    """
    DECIMAL = 'Decimal'
    MINUTES = 'Minutes'
    MINUTES_SECONDS = 'MinutesSeconds'


class DistanceFormat(Enum):
    """
    Enumerations representing the DistanceFormat defined
    in <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
    """
    RE = 'Re'
    KM = 'Km'
    INTEGER_KM = 'IntegerKm'
    SCIENTFIC_NOTATION_KM = 'ScientificNotationKm'


class LatLonFormat(Enum):
    """
    Enumerations representing the LatLonFormat defined
    in <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
    """
    LAT_90_LON_360 = 'Lat90Lon360'
    LAT_90_LON_180 = 'Lat90Lon180'
    LAT_90_SN_LON_180_WE = 'Lat90SnLon180We'


class TimeFormat(Enum):
    """
    Enumerations representing the TimeFormat defined
    in <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
    """
    HH_HHHH = 'hh_hhhh'
    HH_MM_SS = 'hh_mm_ss'
    HH_MM = 'hh_mm'


class FormatOptions:  # pylint: disable=too-many-instance-attributes
    """
    Class representing a FormatOptions type from
    <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

    Properties
    ----------
    date_format
        Date format option.
    time_format
        Time format option.
    distance_format
        Distance format option.
    distance_digits
        Distance digits option.
    degree_format
        Degree format option.
    degree_digits
        Degree digits option.
    lat_lon_format
        Latitude/Longitude format option.
    cdf
        CDF option.
    lines_per_page
        Lines per page option.
    """
    def __init__(
            self,
            date_format: DateFormat = None,
            time_format: TimeFormat = None,
            distance_format: DistanceFormat = None,
            distance_digits: int = None,
            degree_format: DegreeFormat = None,
            degree_digits: DegreeFormat = None,
            lat_lon_format: LatLonFormat = None,
            cdf: bool = None,
            lines_per_page: int = None
        ):  # pylint: disable=too-many-arguments,too-many-branches
        """
        Creates an object representing a FormatOptions type from
        <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

        Parameters
        ----------
        date_format
            Date format option.  If None, default is DateFormat.YYYY_DDD.
        time_format
            Time format option.  If None, default is TimeForamt.HH_HHHH.
        distance_format
            Distance format option.  If None, default is
            DistanceForamt.INTEGER_KM.
        distance_digits
            Distance digits option.  If None, default is 1.
        degree_format
            Degree format option.  If None, default is DegreeFormat.DECIMAL.
        degree_digits
            Degree digits option.  If None, default is 1.
        lat_lon_format
            Latitude/Longitude format option.  If None, default is
            LatLonFormat.LAT_90_LON_360.
        cdf
            CDF option.  If None, default is False.
        lines_per_page
            Lines per page option.  If None, default is 55.
        """

        if date_format is None:
            self._date_format = DateFormat.YYYY_DDD
        else:
            self._date_format = date_format

        if time_format is None:
            self._time_format = TimeFormat.HH_HHHH
        else:
            self._time_format = time_format

        if distance_format is None:
            self._distance_format = DistanceFormat.INTEGER_KM
        else:
            self._distance_format = distance_format

        if distance_digits is None:
            self._distance_digits = 1
        else:
            self._distance_digits = distance_digits

        if degree_format is None:
            self._degree_format = DegreeFormat.DECIMAL
        else:
            self._degree_format = degree_format

        if degree_digits is None:
            self._degree_digits = 1
        else:
            self._degree_digits = degree_digits

        if lat_lon_format is None:
            self._lat_lon_format = LatLonFormat.LAT_90_LON_360
        else:
            self._lat_lon_format = lat_lon_format

        if cdf is None:
            self._cdf = False
        else:
            self._cdf = cdf

        if lines_per_page is None:
            self._lines_per_page = 55
        else:
            self._lines_per_page = lines_per_page


    @property
    def date_format(self):
        """
        Gets the date_format value.

        Returns
        -------
        str
            date_format value.
        """
        return self._date_format


    @date_format.setter
    def date_format(self, value):
        """
        Sets the date_format value.

        Parameters
        ----------
        value
            date_format value.
        """
        self._date_format = value


    @property
    def time_format(self):
        """
        Gets the time_format value.

        Returns
        -------
        str
            time_format value.
        """
        return self._time_format


    @time_format.setter
    def time_format(self, value):
        """
        Sets the time_format value.

        Parameters
        ----------
        value
            time_format value.
        """
        self._time_format = value


    @property
    def distance_format(self):
        """
        Gets the distance_format value.

        Returns
        -------
        str
            distance_format value.
        """
        return self._distance_format


    @distance_format.setter
    def distance_format(self, value):
        """
        Sets the distance_format value.

        Parameters
        ----------
        value
            distance_format value.
        """
        self._distance_format = value


    @property
    def distance_digits(self):
        """
        Gets the distance_digits value.

        Returns
        -------
        str
            distance_digits value.
        """
        return self._distance_digits


    @distance_digits.setter
    def distance_digits(self, value):
        """
        Sets the distance_digits value.

        Parameters
        ----------
        value
            distance_digits value.
        """
        self._distance_digits = value


    @property
    def degree_format(self):
        """
        Gets the degree_format value.

        Returns
        -------
        str
            degree_format value.
        """
        return self._degree_format


    @degree_format.setter
    def degree_format(self, value):
        """
        Sets the degree_format value.

        Parameters
        ----------
        value
            degree_format value.
        """
        self._degree_format = value


    @property
    def degree_digits(self):
        """
        Gets the degree_digits value.

        Returns
        -------
        str
            degree_digits value.
        """
        return self._degree_digits


    @degree_digits.setter
    def degree_digits(self, value):
        """
        Sets the degree_digits value.

        Parameters
        ----------
        value
            degree_digits value.
        """
        self._degree_digits = value


    @property
    def lat_lon_format(self):
        """
        Gets the lat_lon_format value.

        Returns
        -------
        str
            lat_lon_format value.
        """
        return self._lat_lon_format


    @lat_lon_format.setter
    def lat_lon_format(self, value):
        """
        Sets the lat_lon_format value.

        Parameters
        ----------
        value
            lat_lon_format value.
        """
        self._lat_lon_format = value


    @property
    def cdf(self):
        """
        Gets the cdf value.

        Returns
        -------
        str
            cdf value.
        """
        return self._cdf


    @cdf.setter
    def cdf(self, value):
        """
        Sets the cdf value.

        Parameters
        ----------
        value
            cdf value.
        """
        self._cdf = value


    @property
    def lines_per_page(self):
        """
        Gets the lines_per_page value.

        Returns
        -------
        str
            lines_per_page value.
        """
        return self._lines_per_page


    @lines_per_page.setter
    def lines_per_page(self, value):
        """
        Sets the lines_per_page value.

        Parameters
        ----------
        value
            lines_per_page value.
        """
        self._lines_per_page = value


    def xml_element(self) -> ET:
        """
        Produces the XML Element representation of this object.

        Returns
        -------
        ET
            XML Element represenation of this object.
        """

        builder = ET.TreeBuilder()
        builder.start('FormatOptions', {})
        builder.start('DateFormat', {})
        builder.data(self._date_format.value)
        builder.end('DateFormat')
        builder.start('TimeFormat', {})
        builder.data(self._time_format.value)
        builder.end('TimeFormat')
        builder.start('DistanceFormat', {})
        builder.data(self._distance_format.value)
        builder.end('DistanceFormat')
        builder.start('DistanceDigits', {})
        builder.data(str(self._distance_digits))
        builder.end('DistanceDigits')
        builder.start('DegreeFormat', {})
        builder.data(self._degree_format.value)
        builder.end('DegreeFormat')
        builder.start('DegreeDigits', {})
        builder.data(str(self._degree_digits))
        builder.end('DegreeDigits')
        builder.start('LatLonFormat', {})
        builder.data(self._lat_lon_format.value)
        builder.end('LatLonFormat')
        builder.start('Cdf', {})
        builder.data(str(self._cdf).lower())
        builder.end('Cdf')
        builder.start('LinesPerPage', {})
        builder.data(str(self._lines_per_page))
        builder.end('LinesPerPage')
        builder.end('FormatOptions')

        return builder.close()


class CdfFormatOptions(FormatOptions):

    """
    Convenience class representing a FormatOptions object specifying
    CDF format results.
    """
    def __init__(self):
        """
        Constructs a FormatOptions object specifying CDF format results.
        """
        super().__init__(None, None, None, None, None, None, None, 
                         True, None)



#!/usr/bin/env python3

#
# NOSA HEADER START
#
# The contents of this file are subject to the terms of the NASA Open
# Source Agreement (NOSA), Version 1.3 only (the "Agreement").  You may
# not use this file except in compliance with the Agreement.
#
# You can obtain a copy of the agreement at
#   docs/NASA_Open_Source_Agreement_1.3.txt
# or
#   https://sscweb.gsfc.nasa.gov/WebServices/NASA_Open_Source_Agreement_1.3.txt.
#
# See the Agreement for the specific language governing permissions
# and limitations under the Agreement.
#
# When distributing Covered Code, include this NOSA HEADER in each
# file and include the Agreement file at
# docs/NASA_Open_Source_Agreement_1.3.txt.  If applicable, add the
# following below this NOSA HEADER, with the fields enclosed by
# brackets "[]" replaced with your own identifying information:
# Portions Copyright [yyyy] [name of copyright owner]
#
# NOSA HEADER END
#
# Copyright (c) 2013-2020 United States Government as represented by
# the National Aeronautics and Space Administration. No copyright is
# claimed in the United States under Title 17, U.S.Code. All Other
# Rights Reserved.
#

"""
Module defining classes to represent the Result class and its
sub-classes from
<https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.<br>

Copyright &copy; 2013-2020 United States Government as represented by the
National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S.Code. All Other Rights Reserved.
"""

import xml.etree.ElementTree as ET
from datetime import datetime
from typing import Any, Callable, Dict, List
from abc import ABCMeta, abstractmethod
from enum import Enum
import dateutil.parser
import numpy as np

from sscws import ET_NS
from sscws.coordinates import CoordinateSystem
from sscws.regions import FootpointRegion, Hemisphere, SpaceRegion


class ResultStatusCode(Enum):
    """
    Enumerations representing the ResultStatusCode defined
    in <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
    """
    SUCCESS = 'Success'
    CONDITIONAL_SUCCESS = 'ConditionalSuccess'
    ERROR = 'Error'


class ResultStatusSubCode(Enum):
    """
    Enumerations representing the ResultStatusSubCode defined
    in <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
    """
    SUCCESS = 'Success'
    MISSING_REQUEST = 'MissingRequest'
    MISSING_SATELLITES = 'MissingSatellites'
    INVALID_BEGIN_TIME = 'InvalidBeginTime'
    INVALID_END_TIME = 'InvalidEndTime'
    INVALID_SATELLITE = 'InvalidSatellite'
    INVALID_TIME_RANGE = 'InvalidTimeRange'
    INVALID_RESOLUTION_FACTOR = 'InvalidResolutionFactor'
    MISSING_OUTPUT_OPTIONS = 'MissingOutputOptions'
    MISSING_COORD_OPTIONS = 'MissingCoordOptions'
    MISSING_COORD_SYSTEM = 'MissingCoordSystem'
    INVALID_COORD_SYSTEM = 'InvalidCoordSystem'
    MISSING_COORD_COMPONENT = 'MissingCoordComponent'
    MISSING_GRAPH_OPTIONS = 'MissingGraphOptions'
    MISSING_COORDINATE_SYSTEM = 'MissingCoordinateSystem'
    MISSING_COORDINATE_COMPONENT = 'MissingCoordinateComponent'
    SERVER_ERROR = 'ServerError'


class Result(metaclass=ABCMeta):
    """
    Class representing a Result from
    <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

    Notes
    -----
    Although this class is essentially a dictionary, it was defined as a
    class to make certain that it matched the structure and key names
    of a Request from
    <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
    It also needs to function as a base class for the concrete
    sub-classes of a Request.

    Properties
    ----------
    status_code
        Result status code.
    status_sub_code
        Result status sub-code.
    status_text
        Result status text.
    """
    @abstractmethod
    def __init__(
            self,
            status_code: ResultStatusCode,
            status_sub_code: ResultStatusSubCode,
            status_text: List[str]):
        """
        Creates an object representing a Result from
        <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

        Parameters
        ----------
        status_code
            Result status code value.
        status_sub_code
            Result status subcode value.
        status_text
            Status text.
        """
        self._status_code = status_code
        self._status_sub_code = status_sub_code
        self._status_text = status_text


    @property
    def status_code(self):
        """
        Gets the status_code value.

        Returns
        -------
        str
            status_code value.
        """
        return self._status_code


    @status_code.setter
    def status_code(self, value):
        """
        Sets the status_code value.

        Parameters
        ----------
        value
            status_code value.
        """
        self._status_code = value



    @property
    def status_sub_code(self):
        """
        Gets the status_sub_code value.

        Returns
        -------
        str
            status_sub_code value.
        """
        return self._status_sub_code


    @status_sub_code.setter
    def status_sub_code(self, value):
        """
        Sets the status_sub_code value.

        Parameters
        ----------
        value
            status_sub_code value.
        """
        self._status_sub_code = value



    @property
    def status_text(self):
        """
        Gets the status_text value.

        Returns
        -------
        str
            status_text value.
        """
        return self._status_text


    @status_text.setter
    def status_text(self, value):
        """
        Sets the status_text value.

        Parameters
        ----------
        value
            status_text value.
        """
        self._status_text = value


    @staticmethod
    def get_result(
            result_element: ET
        ) -> Dict:
        """
        Produces a Result from the given xml representation of a Result.

        Parameters
        ----------
        result_element
            ElementTree representation of a Result from
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

        Returns
        -------
        Dict
            Dict representation of the given ElementTree Result
            as described in
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

        Raises
        ------
        ValueError
            If the given xml is not a valid XML representation of Result.
        """

        result_type = result_element.get(\
            '{http://www.w3.org/2001/XMLSchema-instance}type')

        if result_type == 'DataResult':
            return Result.get_data_result(result_element)

        if result_type == 'FileResult':
            return Result.get_file_result(result_element)

        raise ValueError('Unrecognized Result type = ' + result_type)


    @staticmethod
    def get_status(
            result_element: ET
        ) -> Dict:
        """
        Produces a Dict representation of a Result with the StatusCode and
        SubStatusCode values from the given xml representation of a Result.

        Parameters
        ----------
        result_element
            ElementTree representation of a Result from
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

        Returns
        -------
        Dict
            Dict representation of the given ElementTree Result
            as described in
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>
            containing the StatusCode and SubStatusCode values.

        Raises
        ------
        ValueError
            If the given xml is not a valid XML representation of a
            DataResult.
        """

        # should these be string or enum values ???
        return {
            'StatusCode': result_element.find(ET_NS + 'StatusCode').text,
            'StatusSubCode': result_element.find(\
                   ET_NS + 'StatusSubCode').text
        }


    @staticmethod
    def get_data_result(
            data_result_element: ET
        ) -> Dict:
        """
        Produces a Dict representation of a DataResult from the given
        xml representation of a DataResult.

        Parameters
        ----------
        data_result_element
            ElementTree representation of a DataResult from
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

        Returns
        -------
        Dict
            Dict representation of the given ElementTree DataResult
            as described in
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

        Raises
        ------
        ValueError
            If the given xml is not a valid XML representation of a
            DataResult.
        """

        result = Result.get_status(data_result_element)

        data_elements = data_result_element.findall(ET_NS + 'Data')
        data_len = len(data_elements)
        if data_len > 0:
            result['Data'] = np.empty(data_len, dtype=object)

            for data_i, data_element in enumerate(data_elements):

                coords_element = data_element.find(ET_NS + 'Coordinates')

                if coords_element is None:
                    # Is this the correct result for this case???
                    result['Data'][data_i] = {
                        'Id': data_element.find(ET_NS + 'Id').text
                    }
                    continue

                result['Data'][data_i] = {
                    'Id': data_element.find(ET_NS + 'Id').text,
                    'Coordinates': {
                        'CoordinateSystem': CoordinateSystem(\
                            coords_element.find(\
                                ET_NS + 'CoordinateSystem').text),
                    },
                }


                for name in ['X', 'Y', 'Z', 'Latitude', 'Longitude',
                             'LocalTime']:
                    result['Data'][data_i]['Coordinates'][name] = \
                        Result._get_data(coords_element, name, float)

                result['Data'][data_i]['Time'] = \
                    Result._get_data(data_element, 'Time', datetime)


                b_trace_data_elements = data_element.findall(\
                       ET_NS + 'BTraceData')
                b_trace_data_len = len(b_trace_data_elements)
                if b_trace_data_len > 0:

                    result['Data'][data_i]['BTraceData'] = \
                        np.empty(b_trace_data_len, dtype=object)

                    for b_trace_data_i, b_trace_data in \
                        enumerate(b_trace_data_elements):

                        result['Data'][data_i]['BTraceData'][b_trace_data_i] = {
                            'CoordinateSystem': CoordinateSystem(\
                                b_trace_data.find(\
                                    ET_NS + 'CoordinateSystem').text),
                            'Hemisphere': Hemisphere(b_trace_data.find(\
                                 ET_NS + 'Hemisphere').text),
                        }
                        for name in ['Latitude', 'Longitude', 'ArcLength']:
                            result['Data'][data_i]['BTraceData'][b_trace_data_i][name] = \
                                Result._get_data(b_trace_data, name, float)


                    for name in ['RadialLength', 'MagneticStrength',
                                 'NeutralSheetDistance', 'BowShockDistance',
                                 'MagnetoPauseDistance', 'DipoleLValue',
                                 'DipoleInvariantLatitude']:
                        result['Data'][data_i][name] = \
                            Result._get_data(data_element, name, float)


                    result['Data'][data_i]['SpacecraftRegion'] = \
                        Result._get_data(data_element, 'SpacecraftRegion',
                                         SpaceRegion)

                    for name in ['BGseX', 'BGseY', 'BGseZ']:
                        result['Data'][data_i][name] = \
                            Result._get_data(data_element, name, float)

                    for name in ['RadialTracedFootpointRegions',
                                 'NorthBTracedFootpointRegions',
                                 'SouthBTracedFootpointRegions']:
                        result['Data'][data_i][name] = \
                            Result._get_data(data_element, name,
                                             FootpointRegion)

        return result


    @staticmethod
    def _get_data(
            data_element: ET,
            name: str,
            value_type: Callable[[str], Any],
        ) -> np.ndarray:
        """
        Produces an np.ndarray(m, dtype=value_type) representation of the
        values of specified element in the given data_element ET.

        Parameters
        ----------
        data_element
            ElementTree representation of a DataResult from
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
        name
            Name of element in data_element to get values of.
        value_type
            Type of values to get.

        Returns
        -------
        np.ndarray
            np.ndarray representation of the specified values from
            data_element.
        """
        data_elements = data_element.findall(ET_NS + name)
        size = len(data_elements)
        if size > 0:
            values = np.empty(size, dtype=value_type)
            for index, value in enumerate(data_elements):
                if value_type is datetime:
                    values[index] = dateutil.parser.parse(value.text)
                else:
                    values[index] = value_type(value.text)
        else:
            values = None

        return values


    @staticmethod
    def get_file_result(
            file_result_element: ET
        ) -> Dict:
        """
        Produces a Dict representation of a FileResult from the given
        xml representation of a FileResult.

        Parameters
        ----------
        file_result_element
            ElementTree representation of a FileResult from
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

        Returns
        -------
        Dict
            Dict representation of the given ElementTree FileResult
            as described in
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

        Raises
        ------
        ValueError
            If the given xml is not a valid XML representation of a
            FileResult.
        """

        result = Result.get_status(file_result_element)
        result['Files'] = []

        for file_description in file_result_element.findall(\
                ET_NS + 'Files'):
            result['Files'].append({
                'Name': file_description.find(ET_NS + 'Name').text,
                'MimeType': file_description.find(ET_NS + 'MimeType').text,
                'Length': int(file_description.find(ET_NS + 'Length').text),
                'LastModified': dateutil.parser.parse(\
                    file_description.find(ET_NS + 'LastModified').text)
            })
        return result


#
# If I continue to return Dict instead of class objects, then the rest of
# this file can be deleted.
#
class FileDescription:
    """
    Class representing a FileDescription from
    <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

    Properties
    ----------
    name
        Name of file (usually a URL).
    mime_type
        MIME type of file.
    length
        Length of file in bytes.
    last_modified
        Time when file was last modified.
    """
    def __init__(
            self,
            name: str,
            mime_type: str,
            length: int,
            last_modified: datetime):
        """
        Creates a FileDescription object.

        Parameters
        ----------
        name
            Name of file (usually a URL).
        mime_type
            MIME type of file.
        length
            Length of file in bytes.
        last_modified
            Time when file was last modified.
        """
        self._name = name
        self._mime_type = mime_type
        self._length = length
        self._last_modified = last_modified

    @property
    def name(self):
        """
        Gets the name value.

        Returns
        -------
        str
            name value.
        """
        return self._name


    @name.setter
    def name(self, value):
        """
        Sets the name value.

        Parameters
        ----------
        value
            name value.
        """
        self._name = value


    @property
    def mime_type(self):
        """
        Gets the mime_type value.

        Returns
        -------
        str
            mime_type value.
        """
        return self._mime_type


    @mime_type.setter
    def mime_type(self, value):
        """
        Sets the mime_type value.

        Parameters
        ----------
        value
            mime_type value.
        """
        self._mime_type = value


    @property
    def length(self):
        """
        Gets the length value.

        Returns
        -------
        str
            length value.
        """
        return self._length


    @length.setter
    def length(self, value):
        """
        Sets the length value.

        Parameters
        ----------
        value
            length value.
        """
        self._length = value


    @property
    def last_modified(self):
        """
        Gets the last_modified value.

        Returns
        -------
        str
            last_modified value.
        """
        return self._last_modified


    @last_modified.setter
    def last_modified(self, value):
        """
        Sets the last_modified value.

        Parameters
        ----------
        value
            last_modified value.
        """
        self._last_modified = value


class FileResult(Result):
    """
    Class representing a FileResult from
    <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

    Properties
    ----------
    files
        References to the files containing the requested data.
    """
    def __init__(
            self,
            status_code: ResultStatusCode,
            status_sub_code: ResultStatusSubCode,
            status_text: List[str],
            files: List[FileDescription]):
        """
        Creates an object representing a FileResult from
        <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

        Parameters
        ----------
        status_code
            Result status code value.
        status_sub_code
            Result status subcode value.
        status_text
            Status text.
        files
            List of files.
        """
        super().__init__(status_code, status_sub_code, status_text)
        self._files = files


    @property
    def files(self):
        """
        Gets the files value.

        Returns
        -------
        str
            files value.
        """
        return self._files


    @files.setter
    def files(self, value):
        """
        Sets the files value.

        Parameters
        ----------
        value
            files value.
        """
        self._files = value

#class CoordinateData:
#class BTraceData:
#class SatelliteData:
#class DataResult(Result):
#



#!/usr/bin/env python3

#
# NOSA HEADER START
#
# The contents of this file are subject to the terms of the NASA Open
# Source Agreement (NOSA), Version 1.3 only (the "Agreement").  You may
# not use this file except in compliance with the Agreement.
#
# You can obtain a copy of the agreement at
#   docs/NASA_Open_Source_Agreement_1.3.txt
# or
#   https://sscweb.gsfc.nasa.gov/WebServices/NASA_Open_Source_Agreement_1.3.txt.
#
# See the Agreement for the specific language governing permissions
# and limitations under the Agreement.
#
# When distributing Covered Code, include this NOSA HEADER in each
# file and include the Agreement file at
# docs/NASA_Open_Source_Agreement_1.3.txt.  If applicable, add the
# following below this NOSA HEADER, with the fields enclosed by
# brackets "[]" replaced with your own identifying information:
# Portions Copyright [yyyy] [name of copyright owner]
#
# NOSA HEADER END
#
# Copyright (c) 2013-2021 United States Government as represented by
# the National Aeronautics and Space Administration. No copyright is
# claimed in the United States under Title 17, U.S.Code. All Other
# Rights Reserved.
#

"""
Module for accessing the Satellite Situation Center (SSC) web services
https://sscweb.gsfc.nasa.gov/WebServices/REST/.
"""

import platform
import xml.etree.ElementTree as ET
import logging
from typing import Dict, List, Tuple, Union
import requests
import dateutil.parser

from sscws import __version__, ET_NS
from sscws.coordinates import CoordinateSystem, CoordinateComponent
from sscws.outputoptions import CoordinateOptions, OutputOptions
from sscws.request import DataRequest, SatelliteSpecification
from sscws.result import Result
from sscws.timeinterval import TimeInterval

#try:
#    import spacepy.datamodel as spdm    # type: ignore
#    SPDM_AVAILABLE = True
#except ImportError:
#    SPDM_AVAILABLE = False


class SscWs:
    """
    Class representing the web service interface to NASA's
    Satelite Situation Center (SSC) <https://sscweb.gsfc.nasa.gov/>.

    Notes
    -----
    The logger used by this class has the class' name (SscWs).  By default,
    it is configured with a NullHandler.  Users of this class may configure
    the logger to aid in diagnosing problems.

    This class is dependent upon xml.etree.ElementTree module which is
    vulnerable to an "exponential entity expansion" and "quadratic blowup
    entity expansion" XML attack.  However, this class only receives XML
    from the (trusted) SSC server so these attacks are not a threat.  See
    the xml.etree.ElementTree "XML vulnerabilities" documentation for
    more details
    <https://docs.python.org/3/library/xml.html#xml-vulnerabilities>.
    """
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    def __init__(
            self,
            endpoint=None,
            timeout=None,
            proxy=None,
            ca_certs=None,
            disable_ssl_certificate_validation=False):
        """
        Creates an object representing the SSC web services.

        Parameters
        ----------
        endpoint
            URL of the SSC web service.  If None, the default is
            'https://sscweb.gsfc.nasa.gov/WS/sscr/2/'.
        timeout
            Number of seconds to wait for a response from the server.
        proxy
            HTTP proxy information.  For example,
            proxies = {
              'http': 'http://10.10.1.10:3128',
              'https': 'http://10.10.1.10:1080',
            }
            Proxy information can also be set with environment variables.
            For example,
            $ export HTTP_PROXY="http://10.10.1.10:3128"
            $ export HTTPS_PROXY="http://10.10.1.10:1080"
        ca_certs
            Path to certificate authority (CA) certificates that will
            override the default bundle.
        disable_ssl_certificate_validation
            Flag indicating whether to validate the SSL certificate.
        """

        self.logger = logging.getLogger(type(self).__name__)
        self.logger.addHandler(logging.NullHandler())

        self.retry_after_time = None

        self.logger.debug('endpoint = %s', endpoint)
        self.logger.debug('ca_certs = %s', ca_certs)
        self.logger.debug('disable_ssl_certificate_validation = %s',
                          disable_ssl_certificate_validation)

        if endpoint is None:
            self._endpoint = 'https://sscweb.gsfc.nasa.gov/WS/sscr/2/'
        else:
            self._endpoint = endpoint
        self._user_agent = 'sscws/' + __version__ + ' (' + \
            platform.python_implementation() + ' ' \
            + platform.python_version() + '; '+ platform.platform() + ')'
        self._request_headers = {
            'Content-Type' : 'application/xml',
            'Accept' : 'application/xml',
            'User-Agent' : self._user_agent
        }
        self._session = requests.Session()
        #self._session.max_redirects = 0
        self._session.headers.update(self._request_headers)

        if ca_certs is not None:
            self._session.verify = ca_certs

        if disable_ssl_certificate_validation is True:
            self._session.verify = False

        if proxy is not None:
            self._proxy = proxy

        self._timeout = timeout

    # pylint: enable=too-many-arguments


    def __str__(self) -> str:
        """
        Produces a string representation of this object.

        Returns
        -------
        str
            A string representation of this object.
        """
        return 'SscWs(endpoint=' + self._endpoint + ', timeout=' + \
               str(self._timeout) + ')'


    def __del__(self):
        """
        Destructor.  Closes all network connections.
        """

        self.close()


    def close(self) -> None:
        """
        Closes any persistent network connections.  Generally, deleting
        this object is sufficient and calling this method is unnecessary.
        """
        self._session.close()


    def get_observatories(
            self
        ) -> Tuple[int, List[Dict]]:
        """
        Gets a description of the available SSC observatories.

        Returns
        -------
        int
            HTTP status code (200 is successful).
        List
            An array of ObservatoryDescription dictionaries where the
            structure of the dictionary mirrors ObservatoryDescription in
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
        """
        url = self._endpoint + 'observatories'

        self.logger.debug('request url = %s', url)

        response = self._session.get(url, timeout=self._timeout)

        if response.status_code != 200:

            self.logger.info('%s failed with http code %d', url,
                             response.status_code)
            self.logger.info('response.text: %s', response.text)
            return (response.status_code, [])

        observatory_response = ET.fromstring(response.text)

        observatories = []

        for observatory in observatory_response.findall(ET_NS + 'Observatory'):

            observatories.append({
                'Id': observatory.find(ET_NS + 'Id').text,
                'Name': observatory.find(ET_NS + 'Name').text,
                'Resolution': int(observatory.find(ET_NS + 'Resolution').text),
                'StartTime': dateutil.parser.parse(observatory.find(\
                    ET_NS + 'StartTime').text),
                'EndTime': dateutil.parser.parse(observatory.find(\
                    ET_NS + 'EndTime').text),
                'ResourceId': observatory.find(ET_NS + 'ResourceId').text
            })

        return (response.status_code, observatories)


    def get_ground_stations(
            self
        ) -> Tuple[int, List[Dict]]:
        """
        Gets a description of the available SSC ground stations.

        Returns
        -------
        int
            HTTP status code (200 is successful).
        List
            An array of GroundStations dictionaries where the
            structure of the dictionary mirrors GroundStations in
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
        """
        url = self._endpoint + 'groundStations'

        self.logger.debug('request url = %s', url)

        response = self._session.get(url, timeout=self._timeout)

        if response.status_code != 200:

            self.logger.info('%s failed with http code %d', url,
                             response.status_code)
            self.logger.info('response.text: %s', response.text)
            return (response.status_code, [])

        ground_station_response = ET.fromstring(response.text)

        ground_stations = []

        for ground_station in ground_station_response.findall(\
                ET_NS + 'GroundStation'):

            location = ground_station.find(ET_NS + 'Location')
            latitude = float(location.find(ET_NS + 'Latitude').text)
            longitude = float(location.find(ET_NS + 'Longitude').text)

            ground_stations.append({
                'Id': ground_station.find(ET_NS + 'Id').text,
                'Name': ground_station.find(ET_NS + 'Name').text,
                'Location': {
                    'Latitude': latitude,
                    'Longitude': longitude
                }
            })

        return (response.status_code, ground_stations)


    def get_locations(
            self,
            param1: Union[List[str], DataRequest],
            time_range: Union[List[str], TimeInterval] = None,
            coords: List[CoordinateSystem] = None
        ) -> Tuple[int, Dict]:
        """
        Gets the specified locations.  Complex requests (requesting
        magnetic field model values) require a single DataRequest
        parameter.  Simple requests (for only x, y, z, lat, lon,
        local_time) require at least the first two paramters.

        Parameters
        ----------
        param1
            A locations DataRequest or a list of observatory identifier
            (returned by get_observatories).
        time_range
            A TimeInterval or two element array of ISO 8601 string
            values of the start and stop time of requested data.
        coords
            Array of CoordinateSystem values that location information
            is to be in.  If None, default is CoordinateSystem.GSE.
        Returns
        -------
        int
            HTTP status code (200 is successful).
        Dict
            Dictionary whose structure mirrors Result from
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
        Raises
        ------
        ValueError
            If param1 is not a DataRequest and time_range is missing or
            time_range does not contain valid values.
        """

        if isinstance(param1, DataRequest):
            request = param1
        else:
            request = SscWs.__create_locations_request(param1, time_range,
                                                       coords)

        status, result = self.__get_locations(request)

        return (status, result)


    #def get_data_from_files(
    #        self,
    #        files: FileResult
    #    ) -> List['spdm.SpaceData']:
    #    """
    #    Gets the given files from the server and returns the contents
    #    in a SpaceData objects.
    #
    #    Parameters
    #    ----------
    #    files
    #        requested files.
    #    Returns
    #    -------
    #    List[SpaceData] ???
    #        The contents of the given files in a SpaceData objects.
    #    """
    #    import spacepy.datamodel as spdm        # type: ignore


    @staticmethod
    def __create_locations_request(
            obs_ids: List[str],
            time_range: Union[List[str], TimeInterval] = None,
            coords: List[CoordinateSystem] = None
        ) -> DataRequest:
        """
        Creates a "simple" (only x, y, z, lat, lon, local_time in GSE)
        locations DataRequest for the given values.
        More complicated requests should be made with DataRequest
        directly.

        Parameters
        ----------
        obs_ids
            A list of observatory identifier (returned by
            get_observatories).
        time_range
            A TimeInterval or two element array of ISO 8601 string
            values of the start and stop time of requested data.
        coords
            Array of CoordinateSystem values that location information
            is to be in.  If None, default is CoordinateSystem.GSE.
        Returns
        -------
        DataRequest
            A simple locations DataRequest based upon the given values.
        Raises
        ------
        ValueError
            If time_range is missing or time_range does not contain
            valid values.
        """

        sats = []
        for sat in obs_ids:
            sats.append(SatelliteSpecification(sat, 1))

        if time_range is None:
            raise ValueError('time_range value is required when ' +
                             '1st is not a DataRequest')

        if isinstance(time_range, list):
            time_interval = TimeInterval(time_range[0], time_range[1])
        else:
            time_interval = time_range

        if coords is None:
            coords = [CoordinateSystem.GSE]

        coord_options = []
        for coord in coords:
            coord_options.append(
                CoordinateOptions(coord, CoordinateComponent.X))
            coord_options.append(
                CoordinateOptions(coord, CoordinateComponent.Y))
            coord_options.append(
                CoordinateOptions(coord, CoordinateComponent.Z))
            coord_options.append(
                CoordinateOptions(coord, CoordinateComponent.LAT))
            coord_options.append(
                CoordinateOptions(coord, CoordinateComponent.LON))
            coord_options.append(
                CoordinateOptions(coord, CoordinateComponent.LOCAL_TIME))

        return DataRequest(None, time_interval, sats, None,
                           OutputOptions(coord_options), None, None)


    def __get_locations(
            self,
            request: DataRequest
        ) -> Tuple[int, Dict]:
        """
        Gets the given locations DataRequest.

        Parameters
        ----------
        request
            A locations DataRequest.
        Returns
        -------
        int
            HTTP status code (200 is sucessful).
        Dict
            Dictionary whose structure mirrors Result from
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
        """
        url = self._endpoint + 'locations'

        self.logger.debug('POST request url = %s', url)

        xml_data_request = request.xml_element()

        #self.logger.debug('request XML = %s', 
        #                  ET.tostring(xml_data_request))

        response = self._session.post(url,
                                      data=ET.tostring(xml_data_request),
                                      timeout=self._timeout)
        if response.status_code != 200:

            try:
                # requires version 3.9
                ET.indent(xml_data_request)
            except AttributeError:
                pass
            self.logger.debug('request XML = %s', 
                              ET.tostring(xml_data_request))

            self.logger.info('%s failed with http code %d', url,
                             response.status_code)
            self.logger.info('response.text: %s', response.text)
            return (response.status_code, {})

        result_element = ET.fromstring(response.text).find(\
                             ET_NS + 'Result')

        return (response.status_code, self.__get_result(result_element))


    def __get_result(
            self,
            result_element: ET
        ) -> Dict:
        """
        Creates a dict representation of a DataResult from an ElementTree
        representation.

        Parameters
        ----------
        result_element
            ElementTree representation of a DataResult from
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.

        Returns
        -------
        Dict
            Dict representation of the given ElementTree DataResult
            as described in
            <https://sscweb.gsfc.nasa.gov/WebServices/REST/SSC.xsd>.
        """

        try:
            # requires version 3.9
            ET.indent(result_element)
        except AttributeError:
            pass
        #self.logger.debug('result_element XML = %s',
        #                  ET.tostring(result_element))

        return Result.get_result(result_element)



#!/usr/bin/env python3

#
# NOSA HEADER START
#
# The contents of this file are subject to the terms of the NASA Open
# Source Agreement (NOSA), Version 1.3 only (the "Agreement").  You may
# not use this file except in compliance with the Agreement.
#
# You can obtain a copy of the agreement at
#   docs/NASA_Open_Source_Agreement_1.3.txt
# or
#   https://sscweb.gsfc.nasa.gov/WebServices/NASA_Open_Source_Agreement_1.3.txt.
#
# See the Agreement for the specific language governing permissions
# and limitations under the Agreement.
#
# When distributing Covered Code, include this NOSA HEADER in each
# file and include the Agreement file at
# docs/NASA_Open_Source_Agreement_1.3.txt.  If applicable, add the
# following below this NOSA HEADER, with the fields enclosed by
# brackets "[]" replaced with your own identifying information:
# Portions Copyright [yyyy] [name of copyright owner]
#
# NOSA HEADER END
#
# Copyright (c) 2020-2021 United States Government as represented by
# the National Aeronautics and Space Administration. No copyright is
# claimed in the United States under Title 17, U.S.Code. All Other
# Rights Reserved.
#

"""
Package for accessing the NASA's Satellite Situation Center (SSC) web 
services https://sscweb.gsfc.nasa.gov/WebServices/REST/.

Copyright &copy; 2020-2021 United States Government as represented by the
National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S.Code. All Other Rights Reserved.

Notes
-----
This library does not currently implement the /graphs and /conjunction
resources.  If there is interest in these, support could be added
in the future.

The core functionality is implemented in the `sscws.sscws` sub-module.
New users should start by viewing the `sscws.sscws` sub-module.
"""


__version__ = "0.1.12"

#
# XML schema namespace
#
NS = 'http://sscweb.gsfc.nasa.gov/schema'
#
# Namespace for use in xml.etree.ElementTree.find*
#
ET_NS = '{' + NS + '}'

